#!/usr/bin/perl
use strict;

#================================================================
# generate.pl
# bmajoros@duke.edu
#
# This script generates synthetic genome data, for toy gene-
# finding examples.  Note that the weight matrices generated do
# not include the consensus region; they cover only the noncoding
# margin next to the consensus bases.
#================================================================

my $MAX_FEATURE_LEN=30;
my $MIN_FEATURE_LEN=5;
my $MAX_NUM_INTRONS=3;
my $PROB_SINGLE_EXON_GENE=0.3;
my $WMM_GEOMETRIC_PARM=0.2; # smaller values create a larger bias
my $CONTEXT_LEN=5; # length of weight matrices next to signals
my $EXPONENT=5;#1.8;
my $NUM_RANDOM_SWAPS=70;#58;
my (%codingBaseProbs,%noncodingBaseProbs,@codonRouletteWheel,
    @noncodingWheel,@stopCodons);

push @stopCodons,"TAG";
push @stopCodons,"TGA";
push @stopCodons,"TAA";

$noncodingBaseProbs{'A'}=0.25;
$noncodingBaseProbs{'T'}=0.25;
$noncodingBaseProbs{'C'}=0.25;
$noncodingBaseProbs{'G'}=0.25;

# These three parameters must be modified together in order to keep the
# coding GC% constant:
my $RANDOMIZATION_FACTOR=0.0000001; #0.4;
$codingBaseProbs{'A'}=$codingBaseProbs{'T'}=0.24445;
$codingBaseProbs{'C'}=$codingBaseProbs{'G'}=0.27555;

die "$0 <filestem> <length>\n" unless @ARGV==2;
my ($filestem,$desiredLength)=@ARGV;

my ($codonRouletteWheelSize,$noncodingWheelSize);

my $trainFasta="$filestem-train.fasta";
my $testFasta="$filestem-test.fasta";
my $trainGFF="$filestem-train.gff";
my $testGFF="$filestem-test.gff";
my $codonFile="$filestem.codons";
my $wmmFile="$filestem.wmm";
open(CODONS,">$codonFile") || die "Can't create $codonFile\n";
open(WMM,">$wmmFile") || die "Can't create $wmmFile\n";

#-----------------------------------------------------------------
# Generate weight matrices for signal contexts
my $donorContext=generateWeightMatrix($CONTEXT_LEN);
my $acceptorContext=generateWeightMatrix($CONTEXT_LEN);
my $startCodonContext=generateWeightMatrix($CONTEXT_LEN);
my $stopCodonContext=generateWeightMatrix($CONTEXT_LEN);
saveWMM($donorContext,\*WMM,"donor-site");
saveWMM($acceptorContext,\*WMM,"acceptor-site");
saveWMM($startCodonContext,\*WMM,"start-codon");
saveWMM($stopCodonContext,\*WMM,"stop-codon");

#-----------------------------------------------------------------
# Generate all 64 triples and assign them biased random probabilities
print "
IF THE CODING GC% VALUES PRINTED BELOW ARE NOT SUBSTANTIALLY
GREATER THAN THE NONCODING GC% VALUES, PLEASE RUN THIS PROGRAM
AGAIN.  A LARGER DIFFERENCE WILL ALLOW GREATER PREDICTION ACCURACY
WHEN USING METHODS BASED ON GC-CONTENT.

";

generate($trainFasta,$trainGFF,"train");
generate($testFasta,$testGFF,"test");


#end of main program.







#-----------------------------------------------------------------
#                        SUBROUTINES
#-----------------------------------------------------------------


sub generate 
{
  my ($fastaFile,$gffFile,$label)=@_;
  open(GFF,">$gffFile") || die "Can't create $gffFile\n";
  #-----------------------------------------------------------------
  # Generate all 64 triples and assign them biased random probabilities
  my @alphabet=('A','T','C','G');
  my @triples;
  for(my $i=0 ; $i<4 ; ++$i)
    {
      my $first=$alphabet[$i];
      for(my $j=0 ; $j<4 ; ++$j)
	{
	  my $second=$alphabet[$j];
	  for(my $k=0 ; $k<4 ; ++$k)
	    {
	      my $third=$alphabet[$k];
	      my $triple="$first$second$third";
	      my $codonScore=
		$codingBaseProbs{$first}*$codingBaseProbs{$second}*
		  $codingBaseProbs{$third};
	      my $randFactor=
		1-$RANDOMIZATION_FACTOR/2+rand($RANDOMIZATION_FACTOR);
	      my $modifiedScore=$codonScore*$randFactor;
	      push @triples,[$triple,$modifiedScore];
	      #print "$triple\t$codonScore $randFactor $modifiedScore\n";
	    }
	}
    }
  #-----------------------------------------------------------------
  # Remove stop codons
  my $n=@triples;
  for(my $i=0 ; $i<$n ; ++$i)
    {
      my $triple=$triples[$i];
      my ($codon,$score)=@$triple;
      my $remove=($codon=~/TAG/ ||
		  $codon=~/TGA/ ||
		  $codon=~/TAA/);
      if($remove) {splice(@triples,$i,1);--$n;--$i;next}
    }
  $n=@triples;
  
  #-----------------------------------------------------------------
  # Randomly swap some pairs, to reduce the GC bias that we introduced above
  for(my $i=0 ; $i<$NUM_RANDOM_SWAPS ; ++$i)
    {
      my $j=int(rand($n));
      my $k=int(rand($n));
      my $tmp=$triples[$j]->[1];
      $triples[$j]->[1]=$triples[$k]->[1];
      $triples[$k]->[1]=$tmp;
    }
  my $sum=0;
  foreach my $triple (@triples)
    {
      if(rand(1)<0.5) {$triple->[1]/=2}
    }
  foreach my $triple (@triples)
    {
      my $modifiedScore=$triple->[1]**$EXPONENT;
      $triple->[1]=$modifiedScore;
      $sum+=$modifiedScore;
    }
  
  #-----------------------------------------------------------------
  # Normalize probabilities
  foreach my $codon (@triples)
    {
      my ($triple,$score)=@$codon;
      if($triple eq "TAG" || $triple eq "TGA" || $triple eq "TAA") 
	{die $triple}
      $score/=$sum;
      $codon->[1]=$score;
      print CODONS "$triple $score\n";
    }
  close(CODONS);
  
  #-----------------------------------------------------------------
  # Compute roulette wheels
  $sum=0;
  foreach my $codon (@triples)
    {
      my ($triple,$score)=@$codon;
      $sum+=$score;
      push @codonRouletteWheel,[$triple,$sum];
    }
  $codonRouletteWheelSize=@codonRouletteWheel;
  $sum=0;
  my @bases=keys %noncodingBaseProbs;
  foreach my $base (@bases)
    {
      my $score=$noncodingBaseProbs{$base};
      $sum+=$score;
      push @noncodingWheel,[$base,$sum];
    }
  $noncodingWheelSize=@noncodingWheel;
  
  #-----------------------------------------------------------------
  # Generate pools of coding and noncoding sequence
  my $codingPool=generateCodons(10000);
  my $noncodingPool=generateNoncoding(30000);
  
  my $gc=($codingPool=~s/([GC])/$1/g);
  $gc/=length($codingPool);
  $gc=int($gc*100+0.5);
  print GFF "# coding GC: $gc%\n";
  print "$label: coding GC = $gc%\n";
  
  $gc=($noncodingPool=~s/([GC])/$1/g);
  $gc/=length($noncodingPool);
  $gc=int($gc*100+0.5);
  print GFF "# noncoding GC: $gc%\n";
  print "$label : noncoding GC = $gc%\n";
  
  #-----------------------------------------------------------------
  # Generate genes
  my $seq="AC";
  my $id=1;
  while(length($seq)<$desiredLength-2*$MAX_FEATURE_LEN)
    {
      # Generate some intergenic
      $seq.=generateNoncoding(generateLen());
      
      # Decide whether single or multi-exon gene
      if(rand(1)<$PROB_SINGLE_EXON_GENE)
	{
	  # Generate a single exon gene
	  $seq.=emitFromWMM($startCodonContext);
	  my $begin=length($seq)+1;
	  my $protein=generateCodons(int(generateLen()/3));
	  my $stop=$stopCodons[int(rand(3))];
	  $seq.="ATG$protein$stop";
	  my $end=length($seq);
	  print GFF 
    "1\ttoy-genome\tsingle-exon\t$begin\t$end\t.\t+\t.\ttransgrp=$id;\n";
	  ++$id;
	  $seq.=emitFromWMM($stopCodonContext);
	}
      else # Multi-exon gene...
	{
	  $seq.=emitFromWMM($startCodonContext);
	  
	  # Generate a protein long enough for the number of exons
	  my $numIntrons=int(rand($MAX_NUM_INTRONS));
	  my $numExons=$numIntrons+1;
	  my $protein=generateCodons($numExons*int(generateLen()/3));
	  
	  # Now insert the introns right-to-left
	  my $begin=length($seq);
	  my $leftmost=10;
	  my $rightmost=length($protein)-10;
	  my @spliceCoords;
	  for(my $i=0 ; $i<$numIntrons && $leftmost<$rightmost ; ++$i)
	    {
	      my $intron="GT".generateNoncoding(generateLen())."AG";
	      my $intronLen=length($intron);
	      imposeWMM($donorContext,\$intron,2);
	      imposeWMM($acceptorContext,\$intron,$intronLen-2-$CONTEXT_LEN);
	      my $pos=$leftmost+int(rand($rightmost-$leftmost));
	      substr($protein,$pos,0)=$intron;
	      $rightmost=$pos-10;

	      my $n=@spliceCoords;
	      for(my $j=0 ; $j<$n ; ++$j) {$spliceCoords[$j]+=$intronLen}
	      unshift @spliceCoords,$begin+3+$pos+$intronLen;#intron end
	      unshift @spliceCoords,$begin+3+$pos;           #intron begin
	    }
	  my $stop=$stopCodons[int(rand(3))];
	  $seq.="ATG$protein$stop";
	  my $end=length($seq);

	  $numExons=@spliceCoords/2+1;
	  my $gffBegin=$begin+1;
	  if($numExons==1)
	    {print GFF
       "1\ttoy-genome\tsingle-exon\t$gffBegin\t$end\t.\t+\t.\ttransgrp=$id;\n"}
	  else
	    {
	      my $x=shift @spliceCoords;
	      print GFF 
       "1\ttoy-genome\tinitial-exon\t$gffBegin\t$x\t.\t+\t.\ttransgrp=$id;\n";
	      for(my $j=0 ; $j<$numExons-2 ; ++$j)
		{
		  my $x=shift @spliceCoords;
		  ++$x;
		  my $y=shift @spliceCoords;
		  print GFF 
       "1\ttoy-genome\tinternal-exon\t$x\t$y\t.\t+\t.\ttransgrp=$id;\n";
		}
	      $x=shift @spliceCoords;
	      ++$x;
	      print GFF 
		"1\ttoy-genome\tfinal-exon\t$x\t$end\t.\t+\t.\ttransgrp=$id;\n";
	    }
	  ++$id;
	  $seq.=emitFromWMM($stopCodonContext);
	}
    }
  $seq.=generateNoncoding(generateLen());
  
  # Produce output
  writeFasta(">1 toy-genome",$seq,$fastaFile,60);
  close(GFF);
}


sub writeFasta
  {
    my ($defline,$sequence,$filename,$width)=@_;
    open(FASTA,">$filename") || die "Can't create $filename";
    chop $defline if($defline=~/\n$/);
    $defline=">$defline" unless $defline=~/^>/;
    print FASTA "$defline\n";
    my $length=length $sequence;
    my $numLines=int($length/$width);
    ++$numLines if($length % $width);
    my $start=0;
    for(my $i=0 ; $i<$numLines ; ++$i)
      {
	my $line=substr($sequence,$start,$width);
	print FASTA "$line\n";
	$start+=$width;
      }
    if($length==0) {print FASTA "\n"}
    close(FASTA);
  }


######################################################################
sub generateCodons
  {
    my ($numCodons)=@_;
    my $seq;
    for(my $i=0 ; $i<$numCodons ; ++$i)
      {
	my $r=rand(1);
	for(my $i=0 ; $i<$codonRouletteWheelSize ; ++$i)
	  {
	    my ($codon,$score)=@{$codonRouletteWheel[$i]};
	    if($codon eq "TAG" || $codon eq "TGA" || $codon eq "TAA") 
	      {die $codon}
	    if($r<$score) {$seq.=$codon;last}
	  }
      }
    return $seq;
  }
sub generateNoncoding
  {
    my ($len)=@_;
    my $seq;
    for(my $i=0 ; $i<$len ; ++$i)
      {
	my $r=rand(1);
	for(my $i=0 ; $i<$noncodingWheelSize ; ++$i)
	  {
	    my ($codon,$score)=@{$noncodingWheel[$i]};
	    if($r<$score) {$seq.=$codon;last}
	  }
      }
    return $seq;
  }
sub generateLen
  {
    my $len=int(rand($MAX_FEATURE_LEN-$MIN_FEATURE_LEN+1)+$MIN_FEATURE_LEN);
    return $len;
  }
sub generateWeightMatrix
  {
    my ($len)=@_;
    my $M=[];
    for(my $i=0 ; $i<$len ; ++$i)
      {
	# Start with a standard geometric (identical every time)
	my $distr=geometricDistr();

	# Now shuffle the alphabet to randomize it
	for(my $j=0 ; $j<4 ; ++$j)
	  {
	    my $k=int(rand(4));
	    my $a=$distr->[$j];
	    $distr->[$j]=$distr->[$k];
	    $distr->[$k]=$a;
	  }

	# Convert the distribution into a roulette wheel
	my $sum=0;
	for(my $j=0 ; $j<4 ; ++$j)
	  {
	    $sum+=$distr->[$j];
	    $distr->[$j]=$sum;
	  }

	$M->[$i]=$distr;
      }
    return $M;
  }
sub geometricDistr
  {
    my $r=$WMM_GEOMETRIC_PARM;
    my $y1=1/(1+$r+$r**2+$r**3);
    my $y2=$r*$y1;
    my $y3=$r*$y2;
    my $y4=$r*$y3;
    return [$y1,$y2,$y3,$y4];
  }
sub printWMM
  {
    my ($M)=@_;
    my $n=@$M;
    for(my $pos=0 ; $pos<$n ; ++$pos)
      {
	my $distr=$M->[$pos];
	my $A=$distr->[0];
	my $C=$distr->[1];
	my $G=$distr->[2];
	my $T=$distr->[3];
	print "A=$A C=$C G=$G T=$T\n";
      }
  }
sub emitFromDistr
  {
    my ($distr)=@_;
    my $x=rand(1);
    my $alphabet="ACGT";
    for(my $i=0 ; $i<4 ; ++$i)
      {
	if($x<=$distr->[$i]) {return substr($alphabet,$i,1)}
      }
    die
  }
sub emitFromWMM
  {
    my ($M)=@_;
    my $seq;
    my $len=@$M;
    for(my $i=0 ; $i<$len ; ++$i)
      {
	my $distr=$M->[$i];
	my $symbol=emitFromDistr($distr);
	$seq.=$symbol;
      }
    return $seq;
  }
sub imposeWMM
  {
    my ($M,$seqPtr,$pos)=@_;
    substr($$seqPtr,$pos,@$M)=emitFromWMM($M);
  }
sub saveWMM
  {
    my ($M,$file,$label)=@_;
    my $len=@$M;
    print $file "$label:\n";
    print $file "$len\n";
    for(my $i=0 ; $i<$len ; ++$i)
      {
	my $distr=$M->[$i];
	my $prevP=0;
	for(my $i=0 ; $i<4 ; ++$i)
	  {
	    my $cumulativeP=$distr->[$i];
	    my $p=$cumulativeP-$prevP;
	    $prevP=$cumulativeP;
	    print $file "$p\t";
	  }
	print $file "\n";
      }
    print $file "\n";
  }


